import argparse
import inspect
import json 

import numpy as np

from abc import ABC, abstractclassmethod
from typing import Any, Union, Callable

from ..metaclasses import AbstractHyperParser



def inspect_signature(func: Callable) -> dict:
    """
    Inspects the signature of a function and returns the 
    parameters of a function that have default values (aka kwargs)

    :param func: The function who's kwargs we want
    :type func: Callable
    :return: a dictionary with the default values of kwargs as values and their
             names as keys
    :rtype: dict
    """
    signature = inspect.signature(func)
    hp_dict = {
        k: v.default
        for k, v in signature.parameters.items()
        if not v.default is inspect.Parameter.empty
    }
    hp_dict.pop('args', None)
    hp_dict.pop('self', None)
    hp_dict.pop('kwargs', None)
    return hp_dict


class HyperParameter(object):
    """
    A default HyperParameter interface that allows bounded optimisation on a 
    classes input variables. The interface is designed specifically for optuna
    but should contain the information required for any generic bounded 
    optimisation implementation. 
    """
    __slots__ = ('default', 'bounds', 'type', 'length')

    def __init__(self,
                 default: Any,
                 bounds: tuple = None,
                 htype: type = None,
                 length: Union[list, int] = None):
        """
        Constructor for the HyperParameter object.

        :param default: The default value for the hyperparameter
        :type default: Any
        :param bounds: The bounds of the hyperparameter for optimisation will 
                       not be optimised if None, defaults to None
        :type bounds: tuple, optional
        :param htype: type of the hyperparameter, if None will be type(default)
                      , defaults to None
        :type htype: type, optional
        :param length: The length of the hyperparameter, defaults to None
        :type length: Union[list, int], optional
        """
        self.default = default
        self.bounds = bounds
        #Might need to consider if type(default) in list/tuple/np.ndarray
        self.type = htype if htype is not None else type(default)
        self._set_length(length)

    def _set_length(self, length: Union[list, int, None]) -> None:
        """
        Determines the length of the hyperparameter if it is not specified

        :param length: the specified length of the hyperparameter
        :type length: Union[list, int, None]
        """
        if length is not None:
            self.length = length
        elif self.type in [list, tuple, np.ndarray]:
            self.length = len(self.default)
        # elif self.type == np.ndarray:
        #     #I may need to make a special type of HP for numpy arrays
        #     self.length = self.default.shape
        else:
            assert length == None
            self.length = 1

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.default}, {self.bounds}, {self.type}, {self.length})'

    __repr__ = __str__


def dict2hpdict(hp_dict: dict) -> dict:
    """
    Converts all values in hp_dict to HyperParameter types.

    :param hp_dict: The dictionary who's values need to be transformed to 
                    HyperParameter.
    :type hp_dict: dict
    :return: A new dictionary where all values have been transformed to 
             HyperParameter
    :rtype: dict
    """
    return {k: HyperParameter(v) for k, v in hp_dict.items()}


class AbstractParametered(ABC, metaclass=AbstractHyperParser):
    """
    The interface required for any class that the user wishes to optimise. This
    is intended for use with the Component interface. As components within
    the RLOS framework are generally expected to require some level of tuning.

    Generates an argparse.ArgumentParser for any class by parsing the parameters
    of the class and any superclass it has. Doing this as part of the metaclass
    of this interface means that generating an Experiment's cli is automated. 
    """
    @classmethod
    def harvest_parameters(cls, hp_dict: dict = {}):
        """
        Generates the dictionary of hyperparameters for the class. 

        :param hp_dict: Any additional hyperparameters one wishes to add to the 
                        class has to be included in the class definition it is 
                        called by AbstractHyperParser at import, defaults to {}
        :type hp_dict: dict, optional
        """
        hp_dict = {**inspect_signature(cls.__init__), **hp_dict}
        cls.hyperparameters = dict2hpdict(hp_dict)
        cls.merge_hypers()
        cls.generate_argparser()

    @classmethod
    def generate_argparser(cls):
        """
        Creates the argparse.ArgumentParser for the class. Automates defining the
        cli input arguments for an Experiment. All cli arguments are in the format
        --{cls.__name__}-{parameter_name}.
        """
        #Will be a parent parser for the experiment so no help
        cls._parser = argparse.ArgumentParser(add_help=None)
        cls.set_up_hyperparameters()
        # unwanted = [None, [], {}, (), '']
        unwanted = [None]
        for k, v in cls.hyperparameters.items():
            #Handle multiple input arguments
            if isinstance(v.default, list) or isinstance(
                    v.default, tuple) or isinstance(v.default, np.ndarray):
                if isinstance(v.default, np.ndarray):
                    #numpy args will defualt to float32
                    typ = np.float32
                elif (v.default == []) or (v.default == ()) or (v.length == 0):
                    typ = None
                else:
                    typ = type(v.default[0])
                    
                cls._parser.add_argument(f'--{cls.__name__}-{k}'.replace(
                    '_', '-'),
                                         nargs='+',
                                         default=v.default,
                                         type=typ,
                                         help='A multi input argument')
            elif isinstance(v.default, dict):
                cls._parser.add_argument(f'--{cls.__name__}-{k}'.replace(
                    '_', '-'),
                                         default=str(v.default).replace("'",'"'),
                                         type=json.loads,
                                         help='Dictionary as string input')
            #Handle single input arguments
            #Gets rid of Nones <- not necessarily good behaviour
            elif v.default not in unwanted:
                typ = v.type
                if typ == bool:
                    typ = int
                cls._parser.add_argument(f'--{cls.__name__}-{k}'.replace(
                    '_', '-'),
                                         default=v.default,
                                         type=typ,
                                         help='A single input argument')

    @classmethod
    def merge_hypers(cls):
        """
        If the class inherits from a superclass that also has hyperparameters
        then those will be added to the hyperparameters of this class. 
        """
        if len(cls.mro()) > 1:
            super_class = cls.mro()[1]
            if hasattr(super_class, 'hyperparameters'):
                super_parameters = super_class.hyperparameters
                for key in super_parameters.keys():
                    if key not in cls.hyperparameters:
                        cls.hyperparameters[key] = super_parameters[key]

    @abstractclassmethod
    def set_up_hyperparameters(cls):
        """
        This function is used to add the bounds to any Component's hyperparameters.
        The optimisation library will parse a Component's hyperparameters and any 
        HyperParameter with a non NoneType bounds will be included in the optimisation
        process.
        """
        pass


def get_class_args(klass: AbstractParametered, args: dict) -> dict:
    """
    Helper function to generate the input arguments for an AbstractParametered
    klass given a dictionary args. args is generated by parsing the
    argparse.ArgumentParser given by an experiment after a program is invoked 
    that contains AbstractParametered components. 

    :param klass: The class who's arguments we require
    :type klass: AbstractParametered
    :param args: A dictionary that contains multiple component's input variables
    :type args: dict
    :return: A dictionary that contains the the input arguments required to create
             an instance of klass
    :rtype: dict
    """
    class_args = {}
    for k, v in args.items():
        if not k.startswith(f'{klass.__name__}_'):
            continue
        hp_k = k.replace(f'{klass.__name__}_', '')
        if hasattr(klass, 'hyperparameters') and hp_k in klass.hyperparameters:
            class_args[hp_k] = v
    return class_args


from .component import AbstractComponent
from .hyperparameter import AbstractParametered, get_class_args, HyperParameter
from .io import MasterIO
from .experiment import AbstractExperiment
from .factory import AbstractComponentFactory


from ..interfaces.abstractbaseclasses.component import AbstractComponent

class ProxyLogger:
    def __init__(self, logger, component_name):
        self._logger = logger
        self._component_name = component_name
        self._component_id = self.get_component_id()

    def __getattr__(self, name):
        return getattr(self._logger, name)
    
    def record(self, tag_name, input_value, *args, **kwargs):
        tag_name = f"{'train' if AbstractComponent._is_training else 'eval'}/{tag_name}"
        self._logger.record(self._component_id, self._component_name, tag_name, input_value)

    def get_component_id(self):
        component_id = self._logger._workers[0].get_component_id(self._component_name)
        return component_id


import argparse
import threading
import time

from ..experimentrunners import LocalRunner

def create_dict_from_exp_file(exp_file):
    params = exp_file.split(' ')[1:]  # Remove the python file bit from the front
    params_dict = {}
    i = 0
    while i < len(params):
        key = params[i]
        if key.startswith('--'):
            values = []
            i += 1
            while i < len(params) and not params[i].startswith('--'):
                values.append(params[i])
                i += 1
            if len(values) == 1:
                params_dict[key] = values[0]
            else:
                params_dict[key] = values
        else:
            i += 1
    return params_dict

def run_local(num, args, unknowns):
    if args['exp_filepath'] == '':
        raise ValueError('Experiment filepath is required')
    #read the experiment file
    with open(args['exp_filepath'], 'r') as f:
        exp_file = f.read().strip()
    params = create_dict_from_exp_file(exp_file)
    for i in range(0, len(unknowns)-1, 2):
        if unknowns[i] in params:
            params[unknowns[i]] = unknowns[i+1]
    params['--name'] = params['--name'] + f'_{num}'
    local_runner = LocalRunner()
    hp_cmd_string = ' '.join([f'{key} {" ".join(params[key]) if isinstance(params[key], list) else params[key]}' for key in params])
    local_runner.run(hp_cmd_string, num)

def main(args, unknowns):
    threads = []
    for i in range(args['num_runs']):
        thread = threading.Thread(target=run_local, args=(i, args.copy(), unknowns.copy()))
        thread.start()
        threads.append(thread)
        time.sleep(20)
    for thread in threads:
        thread.join()        

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--exp-filepath', default='')
    parser.add_argument('--num-runs', type=int, default=5)
    args, unknowns = parser.parse_known_args()
    args = vars(args)
    main(args, unknowns)


from setuptools import setup, find_packages

setup(
    name="algos_core",  
    version="0.1.0",  
    author="LlewynS",  
    author_email="algos251@gmail.com",  
    description="Algorithm OS: An extensible operating system for algorithmic development",
    long_description=open("README.md").read(),  
    long_description_content_type="text/markdown",
    url="https://github.com/LlewynS/algos",  
    packages=find_packages(include=["algos", "algos.*"]),  
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",  
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10", 
    install_requires=[
        "matplotlib==3.8.2",
        "networkx==3.1",
        "numpy==1.23.5",
        "optuna==4.2.1",
        "pandas==1.5.3",
        "pexpect==4.8.0",
        "psycopg2-binary==2.9.9",
        "pysftp==0.2.9",
        "pytest==7.2.2",
        "PyYAML==6.0.2", 
        "scikit-learn==1.3.2",
        "scipy==1.15.2",
        "seaborn==0.13.2",
        "setuptools==65.5.0",
        "shap==0.46.0",
        "SQLAlchemy==2.0.21",
    ],
)


class DuplicateError(Exception):
    pass


class SingleEntity:
    """
    SingleEntity allows only one instance of the class to be created with a 
    name. This restricts the namespace of all entities to be unique.

    All classes that extend SingleEntity must make their own _entities variable
    or they will share a common dictionary with SingleEntity
    """
    _entities = {}

    def __new__(cls, name: str, owner: object, return_instance: bool = False, *args,
                **kwargs):
        """
        Singleton pattern that either errors when a duplicate is 
        created, or it returns the existing instance. Singletons are
        purely created based on their name. 

        :param name: the name of the entity
        :type name: str
        :param return_instance: whether to raise an error if the entity exists
                                or simply return the entity, defaults to False
        :type return_instance: bool, optional
        :raises DuplicateError: if return_instance is false raises an error if the
                                entity already exists
        :return: The entity
        :rtype: SingleEntity
        """
        if name in cls._entities:
            #raise a duplicate error
            raise DuplicateError(f"{cls} already exists")
        else:
            #create the entity
            new = super().__new__(cls)
            #add entity to the entities dictionary
            cls._entities[name] = new
            return new


class HybridSingleEntity:
    """
    SingleEntity allows only one instance of the class to be created with a 
    name. This restricts the namespace of all entities to be unique.

    All classes that extend SingleEntity must make their own _entities variable
    or they will share a common dictionary with SingleEntity
    """
    _entities = {}

    def __new__(cls, name: str, owner: object, *args,
                **kwargs):
        """
        Hybrid singleton pattern that either errors when a duplicate is 
        created, or it returns the existing instance. Singletons are
        purely created based on their name. 

        :param name: the name of the entity
        :type name: str
        :param return_instance: whether to raise an error if the entity exists
                                or simply return the entity, defaults to False
        :type return_instance: bool, optional
        :raises DuplicateError: if return_instance is false raises an error if the
                                entity already exists
        :return: The entity
        :rtype: SingleEntity
        """
        key = (name, owner)
        if key in cls._entities:
            #entity has been created
            return cls._entities[key]
        else:
            #create the entity
            new = super().__new__(cls)
            #add entity to the entities dictionary
            cls._entities[key] = new
            return new


from collections import namedtuple
import yaml
import os
# Get the directory containing the module
module_dir = os.path.dirname(os.path.abspath(__file__))

DataPointInfo = namedtuple('DataPointInfo', ['tag_name', 'int_value',
                                             'component_id', 'component_name','step',
                                             'float_value', 'bool_value', 
                                             'string_value', 'timestamp']
                        )

EvaluationPoint = namedtuple('EvaluationPoint', ['init_position', 
                                                   'goal_value', 
                                                   'reward',
                                                   'success', 
                                                   'step'])

def create_connection_string_from_yaml(yaml_file:str, db_name: str = None)->str:
    """Concerts a yaml file into a connection string

    :param yaml_file: Connection file
    :type yaml_file: str
    :return: The connection string for the SQL database
    :rtype: str
    """
    if os.path.isabs(yaml_file):    
        with open(yaml_file, 'r') as file:
            config = yaml.safe_load(file)

        db_config = config.get('database', {})
        ssl_config = config.get('ssl', {})
        if db_name:
            db_config['dbname'] = db_name

        connection_string = (
            f"{db_config['dbtype']}://"
            f"{db_config['username']}:{db_config['password']}@" if db_config['username'] else '') + (
            f"{db_config['host']}:{db_config['port']}" if db_config['host'] and db_config['port'] else ''
            ) + f"/{db_config['dbname']}"
            
        
        if ssl_config:
            connection_string += (
                f"?sslmode={ssl_config['mode']}&"
                f"sslcert={ssl_config['cert']}&"
                f"sslkey={ssl_config['key']}&"
                f"sslrootcert={ssl_config['rootcert']}"
            )

        return connection_string
    else:
        return ''



import time
import inspect
from functools import wraps
from typing import List, Union, Dict, Callable, Tuple

def find_para_pos_in_args(func: Callable, parameter: str) -> Union[Tuple[int, dict], None]:
    """Find the position of parameter in func's args

    :param func: The function who's parameter position we are finding
    :type func: Callable
    :param parameter: The name of the parameter of interest
    :type parameter: str
    :return: The position of said parameter
    :rtype: int
    """
    signature = inspect.signature(func)
    for i, para in enumerate(signature.parameters.values()):
        if para.name == parameter:
            return i, para.default if para.default != inspect._empty else None

def listify_inputs(check_param: Union[List[str], Dict[str,str]]) -> List[str]:
    """Returns the list of input names 

    :param check_param: Dict or list of inputs 
    :type check_param: Union[List[str], Dict[str,str]]
    :return: input names
    :rtype: List[str]
    """
    if type(check_param) == list:
        return check_param
    if type(check_param) == dict:
        if type(check_param[list(check_param.keys())[0]]) == str:
            return list(check_param.keys())
    out = []
    for key, val in check_param.items():
        if type(val) == dict:
            out = [*out, *list(val.keys())]
        elif type(val) == list:
            out = [*out, *val]
    return out

def check_inputs(check_param: Union[List[str], Dict[str,str]], inputs:List[str], parameter:str):
    """Helper function to verify the check_param contains all of inputs

    :param check_param: The list of values to check against inputs
    :type check_param: List[str]
    :param inputs: inputs check_param must contain
    :type inputs: List[str]
    :param parameter: the parameter of interest
    :type parameter: str
    :raises KeyError: all of inputs must be in check_param
    """
    check_param = listify_inputs(check_param)
    for input in inputs:
        if input not in check_param:
            raise KeyError(f"{parameter} must contain the string: {input}")

def check_parameter_contains(*inputs: List[str], parameter:str="inputs"):
    """Check that the function's parameter contains all of the strings in inputs

    This function is agnostic to whether the parameter is an arg or kwarg.

    :param parameter: expected to be inputs or outputs for components, defaults to "inputs"
    :type parameter: str, optional
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if parameter in kwargs:
                check_inputs(kwargs[parameter], inputs, parameter)
            else:
                check = find_para_pos_in_args(func, parameter)
                if check is None:
                    raise AttributeError(f"{parameter} is not in {func}")
                elif check[1] is None:
                    check_inputs(args[check[0]-1],inputs,parameter)
                else:
                    check_inputs(check[1], inputs, parameter)                    
            return func(self, *args, **kwargs)
        return wrapper
    return decorator

def check_inputs_contains(*inputs):
    """Specifies the parameter of check_parameter_contains to be inputs
    """
    return check_parameter_contains(*inputs, parameter="inputs")

def check_outputs_contains(*inputs):
    """Specifies the parameter of check_parameter_contains to be outputs
    """
    return check_parameter_contains(*inputs, parameter="outputs")

def check_io_map_contains(*inputs):
    """Specifies the parameter of check_parameter_contains to be outputs
    """
    return check_parameter_contains(*inputs, parameter="io_map")


import pathlib
import subprocess
import time
import os
import logging
from typing import List
import optuna
import yaml

from ..optimiser.sftp import PexpectSFTPClient, FilePathGenerator
from ..optimiser.scriptgen import ScriptGenerator, ScriptRunner
from ..optimiser.experimentrunner import ExperimentRunner
from ..logger import get_coverage, create_experiment_direct

class SBatchGenerator(ScriptGenerator):
    _clist = None
    _sftp = None

    def __init__(self, job_name: str, config_path: str = None, config_kwargs: dict = None):
        self.config_path = config_path
        self.job_name = job_name
        self._config_kwargs = config_kwargs

    def create_config_string(self, exp_name:str) -> str:
        with open(self.config_path, 'r') as f:
            config = yaml.safe_load(f)
        config_str = ''
        defaults = config['defaults']
        if self._config_kwargs is not None:
            defaults = {**defaults, **self._config_kwargs}
        defaults['job_name'] = exp_name
        config_str += config['sbatch'].format(**defaults).rstrip() + " "
        return config_str

    def generate_script(self, hp_cmd_str: str, exp_name:str) -> str:
        file_str = self.create_config_string(exp_name)
        return file_str + hp_cmd_str + "\n"

class SBatchRunner(ScriptRunner):
    def __init__(self, filepath: pathlib.Path, generator: SBatchGenerator):
        self.filepath = filepath
        self._generator = generator

    def save_sbatch_file(self, file_name:str, file_str: str):
        if not self.filepath.exists():
            self.filepath.mkdir(parents=True, exist_ok=True)
        file_path = self.filepath / file_name
        with open(file_path, 'w', newline='\n') as f:
            f.write(file_str)
        return file_path

    def run_sbatch_script(self, file_path) -> int:
        process = subprocess.Popen(args=['sbatch',
                                            str(file_path)],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        stdout = str(stdout, encoding='utf-8')
        print("STDOUT: " ,stdout)
        print("STD ERROR: " , str(stderr, encoding='utf-8'))
        return int(stdout.split(' ')[-1].strip())

    def wait_script_complete(self, sbatch_num: int) -> bool:
        process = subprocess.Popen(args=['squeue', '-j',
                                            str(sbatch_num)],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        stdout, _ = process.communicate()
        stdout = [
            x for x in str(stdout, encoding='utf-8').split('\n')
            if x != ''
        ]
        return len(stdout) > 1
        
    def get_script_result(self, exp_name, storage_string:str) -> float:
        return get_coverage(exp_name, storage_string)

    def run(self, hp_cmd_str: str, trial:optuna.Trial, storage_string:str) -> float:
        lst = hp_cmd_str.split(' ')
        exp_name = lst[lst.index('--name') + 1]
        sbatch_str = self._generator.generate_script(hp_cmd_str, exp_name)
        
        file_path = self.save_sbatch_file(exp_name, sbatch_str)
        #don't forget to generate script and add input argument
        sbatch_num = self.run_sbatch_script(file_path)
        last_step = 0
        while self.wait_script_complete(sbatch_num):
            # I should add a trial.report here. The function should take 
            try:
                result, steps = self.get_script_result(exp_name)
                if steps != last_step:
                    trial.report(result, steps)
                    last_step = steps
            except:
                pass
            time.sleep(60)
        val = self.get_script_result(exp_name, storage_string)
        if val is None:
            return val
        return val[0]

class RemoteSBatchSFTPRunner(SBatchRunner):
    def __init__(self, filepath: pathlib.Path, sftp:PexpectSFTPClient, generator: SBatchGenerator, file_path_generator: FilePathGenerator):
        super().__init__(filepath, generator)
        self._sftp = sftp
        self._file_path_generator = file_path_generator

    def save_sbatch_file(self, file_name:str, file_str: str):
        self.local_sh, self.remote_path = self._file_path_generator(file_name)
        super().save_sbatch_file(self.local_sh, file_str)
        if not self._sftp.exists(self.remote_path.parent):
            self._sftp.makedirs(str(self.remote_path.parent))
        self._sftp.put(str(self.local_sh), str(self.remote_path))
        return self.remote_path

    def run_sbatch_script(self, file_path:str) -> int:
        sbatch_str = self._sftp.execute(
                f'sbatch {file_path}')
        retry = 0
        while retry < 5:
            if sbatch_str is None:
                logging.error(f"Failed to run sbatch script {file_path}")
                retry += 1
                time.sleep(60)
            try:
                ret_code = int(str(sbatch_str).split(' ')[-1].strip())
                logging.info(f"Running sbatch script {file_path} with sbatch number {ret_code}")
                return ret_code
            except ValueError as e:
                logging.error(f"Failed to extract sbatch number from {sbatch_str}")
                if "Unable to contact slurm controller" in str(sbatch_str):
                    retry += 1
                    time.sleep(60)
                else:
                    raise e
        if retry == 5:
            raise Exception(f"Failed to run sbatch script {file_path}")
        
    
    def wait_script_complete(self, sbatch_num: int) -> bool:
        cmd = self._sftp.execute(f'squeue -j {sbatch_num}')
        if cmd is None:
            return True
        return len(cmd.split('\n')) > 1
    
class LocalRunner(ExperimentRunner):
    def __init__(self, file_path: pathlib.Path = None, storage_url: str = None):
        self._file_path = file_path
        self._storage_url = storage_url
        self._cmd = f'python "{self._file_path}" '

    def generate_cmd(self, hp_cmd_str: str) -> str:
        return self._cmd + hp_cmd_str


    def spawn_process(self, cmd, stdout_file, stderr_file) -> int:
        print('Executing command: ', cmd)
        process = subprocess.Popen(cmd,
                                    stdout=stdout_file,
                                    stderr=stderr_file,
                                    shell=True,
                                    text=True, 
                                    env=os.environ.copy())
        return process
        
    def get_script_result(self, exp_name) -> float:
        return get_coverage(exp_name, self._storage_url)

    def run(self, hp_cmd_str: str, trial_num:int, _:str) -> float:
        cmd = self.generate_cmd(hp_cmd_str)
        lst = hp_cmd_str.split(' ')
        exp_name = lst[lst.index('--name') + 1]
        fp = f"./localruns/{exp_name}/"
        if not os.path.exists(fp):
            os.makedirs(fp)
        stdout_file = open(f"{fp}/{exp_name}.out", "w")
        stderr_file = open(f"{fp}/{exp_name}.err", "w")
        #write cmd to a cmd file
        with open(f"{fp}/{exp_name}.cmd", "w") as f:
            f.write(cmd)
        #don't forget to generate script and add input argument
        process = self.spawn_process(cmd, stdout_file, stderr_file)
        std_out_reader = open(f"{fp}/{exp_name}.out", "r")
        most_recent_evaluation = 0
        while 1:
            ret_code = process.poll()
            # read stdout_file to see if new evaluations to report
            lines = reversed(std_out_reader.readlines())
            for i in lines:
                if "EVALUATING@" in i:
                    step = int(i.split('@')[1])
                    if step > most_recent_evaluation:
                        most_recent_evaluation = step
                        proxy = None
                        retries = 0
                        while proxy is None:
                            proxy = self.get_script_result(exp_name)
                            if proxy is not None:
                                result, step = proxy
                                trial_num.report(result, step)
                            else:
                                time.sleep(60)
                                retries += 1
                                if retries > 4:
                                    break
            if ret_code is not None:
                break
            time.sleep(20)
        if ret_code != 0:
            if process.stderr is not None:
                stderr_output = process.stderr.read()
            else:
                stderr_output = "Dunno, read the corresponding .err file"
            std_out_reader.close()
            stdout_file.close()
            stderr_file.close()
            raise Exception(f"Experiment {exp_name} failed with error code {ret_code} with error {stderr_output}")
        std_out_reader.close()
        stdout_file.close()
        stderr_file.close()
        results = self.get_script_result(exp_name)
        return results if results is None else results[0]




from sqlalchemy import create_engine, func, case
from sqlalchemy.exc import IntegrityError, OperationalError
from sqlalchemy.orm import sessionmaker
import multiprocessing
from datetime import datetime

from typing import Dict,Any
import pathlib
import re
import numpy as np
import psycopg2
import os
from optuna.storages._rdb.models import StudyModel

from .databaseworker import DBWorker
from .tables import Base, Experiment, Evaluation, ExpMetadata, Tag, Component
from .util import create_connection_string_from_yaml, module_dir, DataPointInfo, EvaluationPoint

from ..interfaces.patterns.singleton import SingletonMeta




class DatabaseLogger(metaclass=SingletonMeta):
    _instance = None
    _is_training = True

    def __init__(self, base_name, 
                 db_url = None, experiment_start_time=None, 
                 num_workers=2, exp_metadata_dict:Dict[str,Any]={}, frequency=10):
        experiment_dir = exp_metadata_dict.get('file_path', None)
        if db_url is None:
            db_url = get_default_storage()
        if db_url is None:
            db_file = f'{experiment_dir}/primary.db' if experiment_dir else 'primary.db'
            if not isinstance(db_file, pathlib.Path):
                db_file = pathlib.Path(db_file)
            if not db_file.exists():
                db_file.parent.mkdir(parents=True, exist_ok=True)
                db_file.touch(exist_ok=True)
            db_url = f'sqlite:///{db_file}'
        elif ".yaml" in db_url:
            db_url = create_connection_string_from_yaml(db_url)
        backup_db_file = f'{experiment_dir}/backup.db' if experiment_dir else 'backup.db'
        if not isinstance(backup_db_file, pathlib.Path):
            backup_db_file = pathlib.Path(backup_db_file)
        if not backup_db_file.exists():
            backup_db_file.parent.mkdir(parents=True, exist_ok=True)
            backup_db_file.touch(exist_ok=True)
        self._backup_db_url = f'sqlite:///{backup_db_file}'
        if not db_url:
            db_url = self._backup_db_url
        manager = multiprocessing.Manager()
        self.queue = manager.Queue()
        self.stop_signal = manager.Event()
        self.db_url = db_url
        self._step = 0
        self._component_map = {}
        dbworker = DBWorker(self.queue, self.stop_signal)
        experiment_id, backup_experiment_id, self.experiment_name = dbworker.create_experiment(self.db_url, self._backup_db_url, base_name, experiment_start_time)
        dbworker.add_exp_metadata(exp_metadata_dict)
        self.db_url = dbworker.db_url
        self._backup_db_url = dbworker.backup_db_url
        self._workers = [dbworker]
        self.frequency = frequency

        self.pool = multiprocessing.Pool(processes=num_workers)
        self.results = []
        for i in range(num_workers):
            if i>0:
                dbworker = DBWorker(self.queue, self.stop_signal)
                dbworker.experiment_id = experiment_id
                dbworker.backup_experiment_id = backup_experiment_id
                dbworker.db_url = self.db_url
                dbworker.backup_db_url = self._backup_db_url
                self._workers.append(dbworker)
            self.results.append(self.pool.apply_async(self._workers[i].run))

    def create_experiment(self, name, description, start_time, end_time=None):
        self.queue.put((Experiment(name=name, description=description, start_time=start_time, end_time=end_time), 'add'))
        
    def create_tag(self, name, description=None):
        self.queue.put((Tag(name=name, description=description), 'add'))
        
    def record(self, component_id, component_name, tag_name, input_value, *args, ignore_frequency=False, **kwargs):
        if (self._step % self.frequency and not ignore_frequency):
            return
        int_value, float_value, bool_value, string_value, timestamp = None, None, None, None, None
        if isinstance(input_value, (int, np.int32, np.int64)):
            int_value = int(input_value)
        elif isinstance(input_value, (float, np.float32, np.float64)):
            float_value = float(input_value)
        elif isinstance(input_value, (bool, np.bool_)):
            bool_value = bool(input_value)
        elif isinstance(input_value, (str, np.str_)):
            string_value = str(input_value)
        else:
            raise TypeError(f'Invalid data type {type(input_value)}')
        timestamp = timestamp or datetime.now()
        self.queue.put((DataPointInfo(tag_name=tag_name, 
                                      int_value=int_value,
                                      component_id = component_id,
                                      component_name=component_name,
                                      float_value=float_value,
                                      bool_value=bool_value, 
                                      string_value=string_value,
                                      step = self._step,
                                      timestamp=timestamp), 
                                      'add_data_point'))
        
    def record_evaluation(self, init_position, goal_value, reward, success, step):
        self.queue.put((EvaluationPoint(init_position=str(init_position), 
                                         goal_value=str(goal_value),
                                         reward=float(reward),
                                         success=bool(success),
                                         step=int(step)), 
                                         'add_experiment_result'))
    
    def set_step(self, value):
        self._step = value

    def close(self):
        print("Closing Logger...")
        self.stop_signal.set()
        print("stop signal set")
        self.pool.close()
        self.pool.join()
        self._workers[0].experiment_end()
        self._instance = None

    def bind(self, component_name):
        return ProxyLogger(self, component_name)
    
    

class ProxyLogger:
    def __init__(self, logger, component_name):
        self._logger = logger
        self._component_name = component_name
        self._component_id = self.get_component_id()

    def __getattr__(self, name):
        return getattr(self._logger, name)
    
    def record(self, tag_name, input_value, *args, **kwargs):
        tag_name = f"{'train' if DatabaseLogger._is_training else 'eval'}/{tag_name}"
        self._logger.record(self._component_id, self._component_name, tag_name, input_value)

    def get_component_id(self):
        component_id = self._logger._workers[0].get_component_id(self._component_name)
        return component_id


def get_coverage(experiment_name: str, db_url:str = None):
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    success_rate_values = []
    if experiment:
        success_counts = (
            session.query(
                    Evaluation.step,
                    func.count().filter(Evaluation.success.is_(True)).label('success_count')
                )
                .filter(Evaluation.experiment_id == experiment.experiment_id)
                .group_by(Evaluation.step)
                .order_by(Evaluation.step)
                .all()
        )
        # Extract step and success rate values
        step_values = [row.step for row in success_counts]
        success_count_values = [row.success_count for row in success_counts]
        total_test_counts = [session.query(Evaluation).filter(Evaluation.experiment_id == experiment.experiment_id, Evaluation.step == step).count() for step in step_values]
        success_rate_values = [float(success_count) / float(total_tests) for success_count, total_tests in zip(success_count_values, total_test_counts)]
    session.close()
    engine.dispose()
    if len(success_rate_values)==0:
        return None
    return success_rate_values[-1], step_values[-1]

def get_coverages(experiment_name: str, db_url:str = None):
    # Create a SQLAlchemy engine and session
    if db_url is None:
        db_url = get_default_storage()
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    # Query for the experiment ID using the experiment name
    experiment = session.query(Experiment).filter(Experiment.name == experiment_name).first()
    success_rate_values = []
    if experiment:
        success_counts = (
            session.query(
                    Evaluation.step,
                    func.count().filter(Evaluation.success.is_(True)).label('success_count')
                )
                .filter(Evaluation.experiment_id == experiment.experiment_id)
                .group_by(Evaluation.step)
                .order_by(Evaluation.step)
                .all()
        )
        # Extract step and success rate values
        step_values = [row.step for row in success_counts]
        success_count_values = [row.success_count for row in success_counts]
        total_test_counts = [session.query(Evaluation).filter(Evaluation.experiment_id == experiment.experiment_id, Evaluation.step == step).count() for step in step_values]
        success_rate_values = [float(success_count) / float(total_tests) for success_count, total_tests in zip(success_count_values, total_test_counts)]
    session.close()
    engine.dispose()
    if len(success_rate_values)==0:
        return None
    return success_rate_values, step_values
    

def get_unique_experiment_name(base_name, session):
    """
    This function checks if an experiment name exists in the database and, if it does,
    increments a number appended to the name by 1.
    """
    # Regular expression to find a number at the end of the name
    regex = re.compile(r'(.*)_(\d+)$')
    
    # Query the database for names like the base name
    similar_names = session.query(Experiment.name).filter(
        Experiment.name.ilike(f"{base_name}%")
    ).all()
    
    max_number = -1
    for (name,) in similar_names:
        match = regex.match(name)
        if match:
            # If the name matches the pattern, update max_number
            prefix, number = match.groups()
            if prefix.lower() == base_name.lower():
                max_number = max(max_number, int(number))
        elif name.lower() == base_name.lower():
            # If the name is exactly base_name, ensure we at least get _1
            max_number = max(max_number, 0)
    
    # If max_number is 0, the base name is unique and can be used as is
    # Otherwise, append _{max_number+1} to make it unique
    return f'{base_name}_0' if max_number < 0 else f"{base_name}_{max_number + 1}"

def create_experiment_direct(name, db_url=None, start_time=None, study_id=None):
    if db_url is None:
        db_url = create_connection_string_from_yaml(f'{module_dir}/config.yaml')
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    session = Session()
    Base.metadata.create_all(engine)
    successfully_created = False   
    while not successfully_created:
        try:
            if start_time is None:
                    start_time = datetime.now()
            name = get_unique_experiment_name(name, session)
            new_experiment = Experiment(name=name, start_time=start_time, study_id=study_id)
            session.add(new_experiment)
            session.commit()
            experiment_id = new_experiment.experiment_id
            successfully_created = True
        except (psycopg2.errors.UniqueViolation, IntegrityError):
            session.rollback()
    session.close()
    engine.dispose()
    return experiment_id, name

def get_default_storage(db_name:str = None) -> str:
    db_url = os.environ.get("ALGOS_DB_URL")
    if db_url:
        return db_url
    # check if the file exists
    file = os.environ.get("ALGOS_DB_FILE")
    if file and not os.path.exists(file):
        raise FileNotFoundError(f"Configuration file {file} does not exist.")
    elif not file:
        return
    return create_connection_string_from_yaml(file, db_name)


from typing import List
import networkx as nx
import matplotlib.pyplot as plt

from ..abstractbaseclasses import AbstractComponent
from ..patterns import DuplicateError
from ..utility import TypedDict


class ComponentCollection:
    """
    A collection of Components. Acts similarly to a dictionary containing keys. 

    """

    def __init__(self, **kwargs):
        """
        Constructor for ComponentCollection
        """
        self.keys = list(kwargs.keys())
        self.dict = TypedDict(AbstractComponent)
        for key in self.keys:
            self.dict[key] = kwargs[key]
        self._master_component = None

    def run(self, **kwargs):
        """
        Starts and joins all components.

        Terminating conditions and all control is now the responsibility of the components. 
        """
        [self[k]._process_io() for k in self.keys]
        [self[k].start() for k in self.keys]
        AbstractComponent._all_instances_started.set()
        [self[k].join() for k in self.keys]

    def __setitem__(self, key: str, value: AbstractComponent):
        """
        Allows ComponentCollection to have values added 
        using the same syntax as a dict. It is also in a case
        invariant.

        :param key: The key for the value
        :type key: str
        :param value: The value for the key
        :type value: AbstractComponent
        :raises TypeError: key must be a string
        :raises TypeError: value must be a AbstractComponent
        """

        if key not in self.keys:
            self.keys.append(key)
            try:
                return self.dict.__setitem__(key.lower(), value)
            except TypeError as e:
                self.keys.pop(-1)
                raise e
        else:
            raise DuplicateError(f"{key} already in {self.keys}")

    def __getitem__(self, key: str) -> AbstractComponent:
        """
        Allows ComponentCollection to get values 
        using the same syntax as a dict. It is also
        case invariant. 
        
        :param key: The key of the component
        :type key: str
        :return: The component corresponding to the key
        :rtype: AbstractComponent
        """
        return self.dict.__getitem__(key.lower())

    def __repr__(self) -> str:
        """repr method of ComponentCollection
        
        :return: the repr of ComponentCollection
        :rtype: str
        """
        dictrepr = dict.__repr__(self.dict)
        return '%s(%s)' % (type(self).__name__, dictrepr)
    
    def drawme(self):
        # Create directed graph
        G = nx.DiGraph()
        for key in self.keys:
            multi_io = self[key].io
            for io in multi_io.io:
                for inp in io.inputs:
                    G.add_node(inp)
                for out in io.outputs:
                    G.add_node(out)
                for inp in io.inputs:
                    for out in io.outputs:
                        G.add_edge(inp, out, method=io.name)

        # Draw graph
        pos = nx.spring_layout(G)
        nx.draw(G, pos, with_labels=True, font_weight='bold')
        edge_labels = nx.get_edge_attributes(G, 'method')
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)
        plt.show()
